==> kadi-0.12.0-py3-none-any.whl: kadi/modules/permissions/utils.py <==
# Copyright 2020 Karlsruhe Institute of Technology
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from flask import current_app

from .models import Permission
from .models import Role
from kadi.ext.db import db
from kadi.lib.db import get_class_by_tablename
from kadi.lib.db import TimestampMixin
from kadi.lib.utils import rgetattr
from kadi.modules.accounts.models import User
from kadi.modules.groups.models import Group


def setup_system_role(role_name):
    """Setup a system role with corresponding global permissions.

    Will create the given system role as defined in ``SYSTEM_ROLES`` in the
    application's configuration as well as all global permissions of the listed objects,
    which have to be specified in a ``Meta.permissions`` attribute in each corresponding
    model.

    **Example:**

    .. code-block:: python3

        class Foo:
            class Meta:
                permissions = {
                    "global_actions": [
                        ("create", "Create objects."),
                        ("read", "Read all objects."),
                    ],
                }

    :param role_name: The name of the system role to initialize.
    :return: The created (or possibly already existing) role object or ``None`` if the
        given role is invalid.
    """
    system_roles = current_app.config["SYSTEM_ROLES"]

    if role_name not in system_roles:
        return None

    role = Role.query.filter_by(name=role_name, object=None, object_id=None).first()

    if role is None:
        role = Role.create(name=role_name)

    for object_name, global_actions in system_roles[role_name].items():
        model = get_class_by_tablename(object_name)
        if model is None:
            continue

        model_actions = rgetattr(model, "Meta.permissions", {}).get(
            "global_actions", []
        )
        model_actions = [action_description[0] for action_description in model_actions]

        for action in global_actions:
            if action in model_actions:
                permission = Permission.query.filter_by(
                    action=action, object=object_name, object_id=None
                ).first()

                if permission is None:
                    permission = Permission.create(action=action, object=object_name)

                if permission not in role.permissions:
                    role.permissions.append(permission)

    return role


def setup_permissions(object_name, object_id):
    """Setup the default permissions of an object.

    The default actions and roles have to be specified in a ``Meta.permissions``
    attribute in each model.

    **Example:**

    .. code-block:: python3

        class Foo:
            class Meta:
                permissions = {
                    "actions": [
                        ("read", "Read this object."),
                        ("update", "Edit this object."),
                    ],
                    "roles": [("admin", ["read", "update"])],
                }

    :param object_name: The type of object the permissions refer to.
    :param object_id: The ID of the object.
    :return: ``True`` if the permissions were set up successfully, ``False`` otherwise.
    """
    model = get_class_by_tablename(object_name)
    if model is None:
        return False

    permissions = {}
    for action, _ in model.Meta.permissions["actions"]:
        permission = Permission.create(
            action=action, object=object_name, object_id=object_id
        )

        permissions[action] = permission

    for name, actions in model.Meta.permissions["roles"]:
        role = Role.create(name=name, object=object_name, object_id=object_id)

        for action in actions:
            role.permissions.append(permissions[action])

    return True


def delete_permissions(object_name, object_id):
    """Delete all permissions of an object.

    :param object_name: The type of object the permissions refer to.
    :param object_id: The ID of the object.
    """
    roles = Role.query.filter(Role.object == object_name, Role.object_id == object_id)

    for role in roles:
        db.session.delete(role)

    permissions = Permission.query.filter(
        Permission.object == object_name, Permission.object_id == object_id
    )

    for permission in permissions:
        db.session.delete(permission)


def add_role(subject, object_name, object_id, role_name, update_timestamp=True):
    """Add an existing role to a user or group.

    :param subject: The user or group.
    :param object_name: The type of object the role refers to.
    :param object_id: The ID of the object.
    :param role_name: The name of the role.
    :param update_timestamp: (optional) Flag indicating whether the timestamp of the
        underlying object should be updated or not. The object needs to implement
        :class:`TimestampMixin` in that case.
    :return: ``True`` if the role was added successfully, ``False`` if the subject
        already has a role related to the given object or if the object does not exist.
    :raises ValueError: If no role with the given arguments exists.
    """
    model = get_class_by_tablename(object_name)
    obj = model.query.get(object_id)

    if model is None or obj is None:
        return False

    roles = subject.roles.filter(
        Role.object == object_name, Role.object_id == object_id
    )

    if roles.count() > 0:
        return False

    role = Role.query.filter_by(
        object=object_name, object_id=object_id, name=role_name
    ).first()

    if not role:
        raise ValueError("Specified role does not exist.")

    subject.roles.append(role)

    if update_timestamp and isinstance(obj, TimestampMixin):
        obj.update_timestamp()

    return True


def remove_role(subject, object_name, object_id, update_timestamp=True):
    """Remove an existing role of a user or group.

    :param subject: The user or group.
    :param object_name: The type of object the role refers to.
    :param object_id: The ID of the object.
    :param update_timestamp: (optional) Flag indicating whether the timestamp of the
        underlying object should be updated or not. The object needs to implement
        :class:`TimestampMixin` in that case.
    :return: ``True`` if the role was removed successfully, ``False`` if there was no
        role to remove or if the object does not exist.
    """
    model = get_class_by_tablename(object_name)
    obj = model.query.get(object_id)

    if model is None or obj is None:
        return False

    roles = subject.roles.filter(
        Role.object == object_name, Role.object_id == object_id
    )

    if roles.count() == 0:
        return False

    # As in certain circumstances (e.g. merging two users) a subject may have different
    # roles, all roles related to the given object will be removed.
    for role in roles:
        subject.roles.remove(role)

    if update_timestamp and isinstance(obj, TimestampMixin):
        obj.update_timestamp()

    return True


def set_system_role(user, system_role):
    """Set an existing system role for a given user.

    :param user: The user to set the system role for.
    :param system_role: The name of the system role to set.
    :return: ``True`` if the system role was set successfully, ``False`` if the given
        system role does not exist.
    """
    new_role = Role.query.filter_by(
        name=system_role, object=None, object_id=None
    ).first()

    if new_role is None:
        return False

    user_roles = user.roles.filter(Role.object.is_(None), Role.object_id.is_(None))
    # As in certain circumstances (e.g. merging two users) a user may have different
    # system roles, all of them will be removed.
    for role in user_roles:
        user.roles.remove(role)

    user.roles.append(new_role)
    return True


def get_user_roles(object_name, object_id=None):
    """Get all users and roles for a specific object or object type.

    :param object_name: The type of the object.
    :param object_id: (optional) The ID of a specific object.
    :return: The users and corresponding roles of the object(s) as query object.
    """
    user_roles_query = (
        db.session.query(User, Role).join(User.roles).filter(Role.object == object_name)
    )

    if object_id:
        user_roles_query = user_roles_query.filter(Role.object_id == object_id)

    return user_roles_query


def get_group_roles(object_name, object_id=None):
    """Get all groups and roles for a specific object or object type.

    Note that inactive groups will be filtered out.

    :param object_name: The type of the object.
    :param object_id: (optional) The ID of a specific object.
    :return: The groups and corresponding roles of the object(s) as query object.
    """
    group_roles_query = (
        db.session.query(Group, Role)
        .join(Group.roles)
        .filter(Role.object == object_name, Group.state == "active")
    )

    if object_id:
        group_roles_query = group_roles_query.filter(Role.object_id == object_id)

    return group_roles_query


def get_object_roles(object_name):
    """Get all possible roles and corresponding permissions of an object type.

    :param object_name: The type of the object.
    :return: A list of dictionaries in the following form:

        .. code-block:: python3

            [
                {
                    "name": "admin",
                    "permissions": [
                        {
                            "action": "read,
                            "description": "Read this resource.",
                        }
                    ]
                }
            ]
    """
    model = get_class_by_tablename(object_name)
    roles = [
        {
            "name": role,
            "permissions": [
                {
                    "action": action,
                    "description": get_action_description(action, object_name),
                }
                for action in actions
            ],
        }
        for role, actions in rgetattr(model, "Meta.permissions", {}).get("roles", [])
    ]

    return roles


def get_action_description(action, object_name):
    """Get the description of an action corresponding to a specific permission.

    :param action: The name of the action.
    :param object_name: The type of the object the action belongs to.
    :return: The description or ``None`` if no suitable action or no model corresponding
        to the object type could be found.
    """
    model = get_class_by_tablename(object_name)
    actions = rgetattr(model, "Meta.permissions", {}).get("actions", [])

    for object_action, description in actions:
        if object_action == action:
            return description

    return None

==> kadi-0.13.0-py3-none-any.whl: kadi/modules/permissions/utils.py <==
# Copyright 2020 Karlsruhe Institute of Technology
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from functools import wraps

from flask import abort
from flask import current_app
from flask_login import current_user
from flask_login import login_required

from .core import has_permission
from .models import Permission
from .models import Role
from kadi.ext.db import db
from kadi.lib.db import get_class_by_tablename
from kadi.lib.utils import rgetattr
from kadi.modules.accounts.models import User
from kadi.modules.groups.models import Group


def permission_required(action, object_name, object_id_identifier, status_code=403):
    """Decorator to add access restrictions to a view function.

    If a user is not authenticated, the decorater will behave the same as Flask-Login's
    ``login_required`` decorator. Uses
    :func:`kadi.modules.permissions.core.has_permission` to check for access permission.
    If the object or object instance to check do not exist, the request will
    automatically get aborted with a 404 status code.

    :param action: The action to check for.
    :param object_name: The type of object.
    :param object_id_identifier: The name of the variable to use as ``object_id``, which
        needs to be part of the keyword arguments of the decorated function. May also be
        ``None``.
    :param status_code: (optional) The status code to use if no permission was granted.
    """

    def decorator(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            object_id = None
            if object_id_identifier is not None:
                object_id = kwargs[object_id_identifier]

            # Always return 404 if the model or object do not exist.
            model = get_class_by_tablename(object_name)
            if model is None or (
                object_id is not None
                and not model.query.filter_by(id=object_id)
                .with_entities(model.id)
                .first()
            ):
                abort(404)

            if not has_permission(
                current_user, action, object_name, object_id=object_id
            ):
                abort(status_code)

            return func(*args, **kwargs)

        return login_required(decorated_view)

    return decorator


def setup_system_role(role_name):
    """Setup a system role with corresponding global permissions.

    Will create the given system role as defined in ``SYSTEM_ROLES`` in the
    application's configuration as well as all global permissions of the listed objects,
    which have to be specified in a ``Meta.permissions`` attribute in each corresponding
    model.

    **Example:**

    .. code-block:: python3

        class Foo:
            class Meta:
                permissions = {
                    "global_actions": [
                        ("create", "Create objects."),
                        ("read", "Read all objects."),
                    ],
                }

    :param role_name: The name of the system role to initialize.
    :return: The created or existing role object.
    :raises ValueError: If the given role is not a configured system role.
    """
    system_roles = current_app.config["SYSTEM_ROLES"]

    if role_name not in system_roles:
        raise ValueError(f"Role '{role_name}' is not a configured system role.")

    role = Role.query.filter_by(name=role_name, object=None, object_id=None).first()

    if role is None:
        role = Role.create(name=role_name)

    for object_name, global_actions in system_roles[role_name].items():
        model = get_class_by_tablename(object_name)

        model_actions = rgetattr(model, "Meta.permissions", {}).get(
            "global_actions", []
        )
        model_actions = [action_description[0] for action_description in model_actions]

        for action in global_actions:
            if action in model_actions:
                permission = Permission.query.filter_by(
                    action=action, object=object_name, object_id=None
                ).first()

                if permission is None:
                    permission = Permission.create(action=action, object=object_name)

                if permission not in role.permissions:
                    role.permissions.append(permission)

    return role


def get_user_roles(object_name, object_id=None):
    """Get all users and roles for a specific object or object type.

    :param object_name: The type of the object.
    :param object_id: (optional) The ID of a specific object.
    :return: The users and corresponding roles of the object(s) as query object.
    """
    user_roles_query = (
        db.session.query(User, Role).join(User.roles).filter(Role.object == object_name)
    )

    if object_id:
        user_roles_query = user_roles_query.filter(Role.object_id == object_id)

    return user_roles_query


def get_group_roles(object_name, object_id=None):
    """Get all groups and roles for a specific object or object type.

    Note that inactive groups will be filtered out.

    :param object_name: The type of the object.
    :param object_id: (optional) The ID of a specific object.
    :return: The groups and corresponding roles of the object(s) as query object.
    """
    group_roles_query = (
        db.session.query(Group, Role)
        .join(Group.roles)
        .filter(Role.object == object_name, Group.state == "active")
    )

    if object_id:
        group_roles_query = group_roles_query.filter(Role.object_id == object_id)

    return group_roles_query


def get_object_roles(object_name):
    """Get all possible roles and corresponding permissions of an object type.

    :param object_name: The type of the object.
    :return: A list of dictionaries in the following form:

        .. code-block:: python3

            [
                {
                    "name": "admin",
                    "permissions": [
                        {
                            "action": "read,
                            "description": "Read this resource.",
                        }
                    ]
                }
            ]
    """
    model = get_class_by_tablename(object_name)
    roles = [
        {
            "name": role,
            "permissions": [
                {
                    "action": action,
                    "description": get_action_description(action, object_name),
                }
                for action in actions
            ],
        }
        for role, actions in rgetattr(model, "Meta.permissions", {}).get("roles", [])
    ]

    return roles


def get_action_description(action, object_name):
    """Get the description of an action corresponding to a specific permission.

    :param action: The name of the action.
    :param object_name: The type of the object the action belongs to.
    :return: The description or ``None`` if no suitable action or no model corresponding
        to the object type could be found.
    """
    model = get_class_by_tablename(object_name)
    actions = rgetattr(model, "Meta.permissions", {}).get("actions", [])

    for object_action, description in actions:
        if object_action == action:
            return description

    return None
